-- Some other libraries or additions used in this project can have
-- theirs own licences.
--
-- Copyright (C) 2024 wlkmanist <vlad.king5555@mail.ru> - All Rights Reserved.
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms and conditions of the GNU General Public License, version 3,
-- as published by the Free Software Foundation.
--
-- Contact me directly if you want to use the code below in your project some way.
-- Feel free to use this code without any restrictions if you are BeamNG developer.
--
-- This Lua script controlls dynamically wastegate target pressure by rpm using
-- vanilla turbocharger code and adds custom torque curve for ui app.

local M = {}

local floor = math.floor
local max = math.max
local min = math.min
--local sqrt = math.sqrt

--local rpmToAV = 0.10471971768
--local avToRPM = 9.5493
--local invPascalToPSI = 0.00014503773773
--local psiToPascal = 6894.757293178
local torqueToPower = 0.0001404345295653085

local torqueCurveManager = nil
local engine = nil
local wastegateLimitCurve = nil

-- Variables
local isEnabled = true
--local torqueDataUpdatePending = true


local function setTurbochargerWastegateOffset(device, wastegateOffset)
  if not device or not device.turbocharger or not device.turbocharger.isExisting or not device.turbocharger.setWastegateOffset then
    return
  end

  if wastegateOffset == nil or type(wastegateOffset) == "boolean" then  -- Unknown issue handle, for furter debugging
    log("E", "advanced_internals.TurbochargerByRPM", "Unexpected error, should not happen")
    --wastegateOffset = 0
    --isEnabled = false
    return
  end

  device.turbocharger.setWastegateOffset(min(wastegateOffset, 0))
end

local function updateGFX(dt)
  if isEnabled and engine and electrics and electrics.values then
    local rpm = max(electrics.values["rpm"], 0)                         -- get actual engine rpm
    
    if type(rpm) == "number" then    
      setTurbochargerWastegateOffset(engine, wastegateLimitCurve[floor(rpm)])
    end
    
    --if torqueDataUpdatePending == true and torqueCurveManager then    -- registering custom curve again, graph may be unstable without this
    --  torqueCurveManager.sendTorqueData(engine)
    --  torqueDataUpdatePending = false
    --end
  end
end

local function reset()
  setTurbochargerWastegateOffset(engine, 0)                             -- to make sure this code will not brake turbo behavour
end

local function init(jbeamData)                                          -- load arguments from jbeam
  log("D", "advanced_internals.TurbochargerByRPM", "Init")

  if jbeamData.isEnabled ~= nil then
    isEnabled = jbeamData.isEnabled
  end
  
  extensions.load("torqueCurveManager")
  torqueCurveManager = extensions.torqueCurveManager
  if torqueCurveManager then
    log("D", "advanced_internals.TurbochargerByRPM", "torqueCurveManager extension found")
  else
    log("E", "advanced_internals.TurbochargerByRPM", "torqueCurveManager extension not found, TurbochargerByRPM won't work!")
    return
  end

  local engineName = nil
  if jbeamData.engineName ~= nil then
    engineName = jbeamData.engineName
  else
    engineName = "mainEngine"
  end

  --engine = torqueCurveManager.getDevice()
  if powertrain then
    engine = powertrain.getDevice(engineName)
  end
  
  if engine then
    isEnabled = isEnabled
    if torqueCurveManager.getDevice() == nil then
      torqueCurveManager.setDevice(engine)                              -- make sure that torqueCurveManager will work, it is enabled after some lua like this setting the device
    end
  else
    log("E", "advanced_internals.TurbochargerByRPM", "engine == nil, TurbochargerByRPM disabled")
    isEnabled = false
  end
  
  if isEnabled then
    if jbeamData.wastegateOffsetRPM then
      local wastegateCount = #jbeamData.wastegateOffsetRPM
      local tpoints = table.new(wastegateCount, 0)
    
      for i = 1, wastegateCount do                                      -- load "wastegateOffsetRPM" section from jbeam
        local point = jbeamData.wastegateOffsetRPM[i]
        tpoints[i] = {point[1], point[2]}                               -- [ rpm, pressure_offset ],
      end
      
      table.sort(tpoints, M.sortFirst)                                  -- sort by rpm to make this code working with higher level of user stupidity 
      
      wastegateLimitCurve = createCurve(tpoints, true)                  -- actual wastegate offset curve that we can use later
      log("D", "advanced_internals.TurbochargerByRPM", "Curve ready")
    else
      log("E", "advanced_internals.TurbochargerByRPM", "No wastegate limit table found!")
      isEnabled = false
      return                                                            -- no data to work with, exit then
    end
  end
  
  if isEnabled then
    torqueCurveManager.addTorqueDataModifier(M.updateTorqueCoefs)       -- registering ours custom torque curve modifier
    torqueCurveManager.sendTorqueData(engine)                           -- call it at least once, but most likely this function will be called more than a few times from vanilla lua scripts
    M.updateGFX = updateGFX
  end
end

local function sortFirst(a,b)                                           -- sort table by rpm function for table.sort()
  if a[1] < b[1] then
    return true
  end
  return false
end

local function updateTorqueCoefs(data)                                  -- recalculate data using dynamic wastegate threshold for ui apps that using static data like Torque Curve app
  -- we can't know the actual wastegate limit for sure since it's a feedback loop with the pressure, so we just estimate it.
  -- lower wastegate ranges lead to more accurate results.
  
  local torqueData
  
  if not data then
    torqueData = deepcopy(engine.torqueData)
  else
    torqueData = deepcopy(data)
  end
  
  local torqueDataCurves = deepcopy(torqueData.curves)
  local iTurbo = nil
  local iNA = nil
  local curveCounter = 0

  for i in pairs(torqueDataCurves) do
    if torqueDataCurves[i].name == "Turbo" then
      iTurbo = i
    elseif torqueDataCurves[i].name == "NA" then
      iNA = i
    elseif torqueDataCurves[i].name == "Turbo + Offset" then            -- exit if ours patch is already applied
      log("W", "advanced_internals.TurbochargerByRPM", "Turbo + Offset is already existing")
      log("W", "advanced_internals.TurbochargerByRPM", "That can't happen")
      return data
    end
    curveCounter = i
  end
  
  log("D", "advanced_internals.TurbochargerByRPM", "Updating torqueData...")
  
  if iNA == nil then
    log("W", "advanced_internals.TurbochargerByRPM", "No valid 'NA' torqueData curve!")
    return data
  end
  
  if iTurbo == nil then
    log("I", "advanced_internals.TurbochargerByRPM", "No valid 'Turbo' torqueData curve!")
    return data
  end
  
  torqueDataCurves[iTurbo].name = "Turbo + Offset"                      -- new curve name
  torqueDataCurves[iTurbo].priority = 32                                -- hardcode priority for new torqueData curve
  torqueData.maxTorque = 0                                              -- clear values before recalculation
  torqueData.maxTorqueRPM = 0
  torqueData.maxPower = 0
  torqueData.maxPowerRPM = 0
  
  local estimatedWastegateTarget = electrics.values.turboBoostMax - 0.5 -- values in psi, an average between wastegateStart and (+ wastegateRange), range is usualy ~1.0 psi, sadly we can't get range value from turbo.
                                                                        -- we don't actually need precise values, we need ratios to calculate all the "Turbo" curve backwards, because again we can't get values from vanilla turbo lua.
  for i in pairs(torqueDataCurves[iTurbo].torque) do                    -- all the calculation magic below, we using both curves to understand what does only turbocharger to the torque.
    if torqueDataCurves[iTurbo].torque[i] ~= torqueDataCurves[iNA].torque[i] then           -- modifying "Turbo" curve only if it differs to "NA"
      torqueDataCurves[iTurbo].torque[i] = (torqueDataCurves[iTurbo].torque[i] - torqueDataCurves[iNA].torque[i]) * ((max(estimatedWastegateTarget + min(wastegateLimitCurve[i] or 0, 0), 0)) / estimatedWastegateTarget) + torqueDataCurves[iNA].torque[i]  -- do a simple correction for torque table
      torqueDataCurves[iTurbo].power[i] = torqueDataCurves[iTurbo].torque[i] * (i - 1) * torqueToPower    -- torque * rpm * coeff
    end
    
    if torqueDataCurves[iTurbo].torque[i] > torqueData.maxTorque then   -- writing new torque/power/rpm peak values, and even if there is no difference between original curve at this [i] rpm step
      torqueData.maxTorque = torqueDataCurves[iTurbo].torque[i]
      torqueData.maxTorqueRPM = i
    end
    if torqueDataCurves[iTurbo].power[i] > torqueData.maxPower then
      torqueData.maxPower = torqueDataCurves[iTurbo].power[i]
      torqueData.maxPowerRPM = i
    end
  end

  curveCounter = curveCounter + 1
  table.insert(torqueData.curves, curveCounter, torqueDataCurves[iTurbo])   -- adding ours calculated curve to the buffer
  
  table.sort(                                                           -- to make sure .priority affects in real
    torqueData.curves,
    function(a, b)
      local ra, rb = a.priority, b.priority
      if ra == rb then
        return a.name < b.name
      else
        return ra > rb
      end
    end
  )
  
  local dashes = {nil, {10, 4}, {8, 3, 4, 3}, {6, 3, 2, 3}, {5, 3}}     -- dashify the line, graph visuals parameters, we can experiment on that
  for k, v in ipairs(torqueData.curves) do
    v.dash = dashes[k]
    v.width = 2                                                         -- line thickness
  end
  
  engine.torqueData = deepcopy(torqueData)                              -- force apply to the engine values, unnecessary for ui app, but maybe something else using this curves in powertrain like transmissions
  
  --dumpToFile("electrics_test_output.json", electrics) -- DEBUG
  --dumpToFile("enginecustom_test_output.json", engine) -- DEBUG
  --dumpToFile("torqueDatacustom_test_output.json", torqueData) -- DEBUG
  
  return torqueData
end


-- public interface
M.init = init
M.reset = reset
M.updateGFX = nop
M.sortFirst = sortFirst
M.updateTorqueCoefs = updateTorqueCoefs
M.sendTorqueData = sendTorqueData

return M
